from aws_glue_interactive_sessions_kernel.glue_kernel_configs.ConfigRetriever import ConfigRetriever


class OverriddenConfigRetriever(ConfigRetriever):
    def __init__(self, overridden_configs):
        self.overridden_configs = overridden_configs

    def get_config_variable(self, config_name):
        return getattr(self.overridden_configs, config_name)



import ast
import base64
import json
from ipykernel.ipkernel import IPythonKernel
from IPython.display import publish_display_data

from importlib_metadata import version
import os
import sys
import uuid
import requests
import subprocess

from aws_glue_interactive_sessions_kernel.glue_kernel_configs.AWSFileConfigRetriever import AWSFileConfigRetriever
from aws_glue_interactive_sessions_kernel.glue_kernel_configs.DefaultConfigRetriever import DefaultConfigRetriever
from aws_glue_interactive_sessions_kernel.glue_kernel_configs.GlueConfigManager import GlueConfigManager
from aws_glue_interactive_sessions_kernel.glue_kernel_configs.OverriddenConfigRetriever import OverriddenConfigRetriever
from aws_glue_interactive_sessions_kernel.glue_kernel_configs.SystemEnvConfigRetriever import SystemEnvConfigRetriever
from aws_glue_interactive_sessions_kernel.glue_kernel_configs.config import Config, default_configs

try:
    from asyncio import Future
except ImportError:

    class Future(object):
        """A class nothing will use."""

import time
import traceback
from collections import defaultdict

import botocore

from ..glue_kernel_utils.GlueSessionsConstants import *
from hdijupyterutils.ipythondisplay import IpythonDisplay
from IPython import get_ipython
from ..glue_kernel_utils.KernelGateway import KernelGateway
from ..glue_kernel_utils.KernelMagics import KernelMagics
from collections import OrderedDict


class BaseKernel(IPythonKernel):
    time_out = float("inf")
    session_id = None
    new_session_id = None
    session_id_prefix = None
    request_origin = None
    ipython_display = None

    def __init__(self, **kwargs):
        super(BaseKernel, self).__init__(**kwargs)
        self.execution_counter = 1
        self._set_default_arguments()
        self.session_name = "AssumeRoleSession"
        self.should_print_startup_text = True

        if not self.ipython_display:
            self.ipython_display = IpythonDisplay()
        self.kernel_gateway = KernelGateway(self.ipython_display)
        self._register_magics()
        self._setup_terminal_logging()

        # Setup environment variables
        os_env_request_origin = self._retrieve_os_env_variable(REQUEST_ORIGIN)
        if os_env_request_origin:
            self.set_request_origin(os_env_request_origin)
        self.overrides = Config()
        self.retrievers = self._initialize_retrievers(self.overrides)
        self.config_manager = GlueConfigManager(self.retrievers)
        # Once the initialization is complete, check if there is a sessionId to be used.
        os_env_session_id = self._retrieve_os_env_variable(SESSION_ID)
        if os_env_session_id:
            self.set_new_session_id(os_env_session_id)

    def _set_default_arguments(self):
        self.default_arguments = {
            "--glue_kernel_version": self._get_current_kernel_version("aws-glue-sessions"),
            "--enable-glue-datacatalog": "true",
        }

    def display_result(self, statement_output):
        if 'Data' in statement_output:
            statement_output_data = statement_output['Data']

            # The default case would be a text/plain response, but an image result may also be accompanied by a text output.
            if MimeTypes.TextPlain.name in statement_output_data and len(statement_output_data) == 1:
                # The value for dictionary key in the result is: text/plain
                self._send_output(statement_output_data[MimeTypes.TextPlain.name])
            elif 'MimeType' in statement_output_data:
                result = statement_output_data['Result']
                mime_type = statement_output_data['MimeType']
                stdout_text = None
                stderr_text = None

                # Unpack the inline vs. indirect std streams
                if mime_type == MimeTypes.S3URI.value:
                    stdout_text = self.kernel_gateway.get_results(statement_output_data['StdOut']) if 'StdOut' in statement_output_data else ""
                    stderr_text = self.kernel_gateway.get_results(statement_output_data['StdErr']) if 'StdErr' in statement_output_data else ""
                else:
                    stdout_text = statement_output_data['StdOut'] if 'StdOut' in statement_output_data else ""
                    stderr_text = statement_output_data['StdErr'] if 'StdErr' in statement_output_data else ""

                if mime_type == MimeTypes.S3URI.value:
                    response = self.kernel_gateway.get_results(statement_output_data['Result'])
                    result = json.loads(response)
                    if len(result) == 0:
                        # Semantically this case corresponds to the 'stream' result,
                        # which is is meant to be "send_output'd" and actually
                        # arrives on stdout which is headed there anyway.
                        result = None
                        mime_type = None
                    elif result.get(MimeTypes.ImagePng.value):
                        result = result.get(MimeTypes.ImagePng.value)
                        mime_type = MimeTypes.ImagePng.value
                    elif result.get(MimeTypes.TextPlain.value):
                        result = result.get(MimeTypes.TextPlain.value)
                        mime_type = MimeTypes.TextPlain.value

                if mime_type and ';' in mime_type:
                    mime_type_list = mime_type.replace(' ','').split(';')
                    if 'base64' in mime_type_list:
                        result = base64.b64decode(str(result))
                        mime_type_list.remove('base64')
                        mime_type = ';'.join(mime_type_list)

                # Dispatch of results for display
                if stdout_text:
                    self._send_output(stdout_text)
                if stderr_text:
                    self._send_error_output(stderr_text)

                if result and mime_type:
                    display_data = {
                        mime_type: result
                    }
                    publish_display_data(data=display_data, metadata={mime_type: {"width": 640, "height": 480}})

    async def do_execute(
            self, code: str, silent: bool, store_history=True, user_expressions=None, allow_stdin=False
    ):
        # VSCode silently executes few statements that are not giving customers opportunity to configure their session.
        # All those statements have store_history flag set to False.
        # Ignoring those statements when a session is not yet created allowing customers to configure their session.
        if not store_history and (
            not self.get_session_id()
                or self.get_current_session_status() in UNHEALTHY_SESSION_STATUS):
            return await self._complete_cell()
        # Print help text upon startup
        if self.should_print_startup_text:
            await self._print_startup_text()
        code = await self._execute_magics(
            code, silent, store_history, user_expressions, allow_stdin
        )
        statement_id = None

        if not code:
            return await self._complete_cell()

        try:
            if (
                    not self.get_session_id()
                    or self.get_current_session_status() in UNHEALTHY_SESSION_STATUS
            ):
                self.kernel_gateway.initialize_clients(profile=self.get_profile(),
                                                       region=self.get_region(), endpoint_url=self.get_endpoint_url(),
                                                       temp_credentials=self.get_temp_credentials())
                self.create_session()
                if self.get_current_session_status() in UNHEALTHY_SESSION_STATUS:
                    self._send_error_output(f'Current session is in an unhealthy terminal state. If you wish to create a new session use %stop_session before executing any further statements.')
                    return await self._complete_cell()
        except Exception as e:
            if RAY_INVALID_REGION_ERROR in (e.args[0]):
                self._send_output(Colour.Red + Colour.Bold + f'Glue Ray session type is not available in selected region.')
                return await self._complete_cell()
            else:
                sys.stderr.write(f"Exception encountered while creating session: {e} \n")
                self._print_traceback(e)
                return await self._complete_cell()

        try:
            # Run statement
            statement_id = self.kernel_gateway.run_statement(
                self.get_session_id(), code
            )["Id"]
            start_time = time.time()

            try:
                while time.time() - start_time <= self.time_out:
                    statement = self.kernel_gateway.get_statement(self.get_session_id(), statement_id)["Statement"]
                    if statement["State"] in FINAL_STATEMENT_STATUS:
                        return self._construct_reply_content(statement)

                    time.sleep(WAIT_TIME)

                sys.stderr.write(f"Timeout occurred with statement (statement_id={statement_id})")

            except KeyboardInterrupt:
                self._send_output(
                    f"Execution Interrupted. Attempting to cancel the statement (statement_id={statement_id})"
                )
                self.kernel_gateway.cancel_statement(self.get_session_id(), statement_id)
        except botocore.exceptions.ClientError as e:
            error_code = e.response['Error']['Code']
            error_message = e.response['Error']['Message']

            if error_code in 'InvalidInputException':
                if self.get_current_session_status() == TIMEOUT_SESSION_STATUS:
                    self._send_error_output(
                        f"session_id={self.get_session_id()} has reached {self.get_current_session_status()} status. ")
                    self._send_error_output(
                        f"Please re-run the same cell to restart the session. You may also need to re-run "
                        f"previous cells if trying to use pre-defined variables.")
                    self.reset_kernel()
                return await self._complete_cell()

            else:
                self._send_error_output(error_message)
                self._cancel_statement(statement_id)
                return await self._complete_cell()

        except Exception as e:
            sys.stderr.write(f"Exception encountered while running statement: {e} \n")
            self._print_traceback(e)
            self.kernel_gateway.cancel_statement(self.get_session_id(), statement_id)
            return await self._complete_cell()

    def set_temporary_credentials(self, aws_access_key_id, aws_secret_access_key, aws_session_token):
        temp_credentials = defaultdict()
        temp_credentials['temp_aws_access_key_id'] = aws_access_key_id
        temp_credentials['temp_aws_secret_access_key'] = aws_secret_access_key
        temp_credentials['temp_aws_session_token'] = aws_session_token
        setattr(self.overrides, TEMP_CREDENTIALS, temp_credentials)

    def get_temp_credentials(self):
        return self.config_manager.get_config_by_priority(TEMP_CREDENTIALS)

    def configure(self, configs_json):
        kernel_managed_params = {
            "profile",
            "endpoint",
            "region",
            "iam_role",
            "session_id",
            "max_capacity",
            "number_of_workers",
            "worker_type",
            "connections",
            "glue_version",
            "security_config",
            "idle_timeout",
            "tags"
        }
        try:
            configs = ast.literal_eval(configs_json)
            if "profile" in configs:
                self.set_profile(configs.get("profile"))
            if "endpoint" in configs:
                self.set_endpoint_url(configs.get("endpoint"))
            if "region" in configs:
                self.set_region(configs.get("region"))
            if "iam_role" in configs:
                self.set_glue_role_arn(configs.get("iam_role"))
            if "session_id" in configs:
                self.set_new_session_id(configs.get("session_id"))
            if "max_capacity" in configs:
                self.set_max_capacity(configs.get("max_capacity"))
            if "number_of_workers" in configs:
                self.set_number_of_workers(configs.get("number_of_workers"))
            if "worker_type" in configs:
                self.set_worker_type(configs.get("worker_type"))
            if "extra_jars" in configs:
                self.set_extra_jars(configs.get("extra_jars"))
            if "connections" in configs:
                self.set_connections(configs.get("connections"))
            if "security_config" in configs:
                self.set_security_config(configs.get("security_config"))
            if "glue_version" in configs:
                self.set_glue_version(configs.get("glue_version"))
            if "idle_timeout" in configs:
                self.set_idle_timeout(configs.get("idle_timeout"))
            if "tags" in configs:
                self.add_tag_to_resource(configs.get("tags"))
            for arg, val in configs.items():
                if arg not in kernel_managed_params:
                    self.add_default_argument(arg, val)
        except Exception as e:
            sys.stderr.write(
                f"The following exception was encountered while parsing the configurations provided: {e} \n"
            )
            self._print_traceback(e)
            return
        if not configs:
            sys.stderr.write("No configuration values were provided.")
        else:
            self._send_output(f"The following configurations have been updated: {configs}")

    def do_shutdown(self, restart):
        self.stop_session()
        return self._do_shutdown(restart)

    def _do_shutdown(self, restart):
        return super(BaseKernel, self).do_shutdown(restart)

    def set_profile(self, profile):
        setattr(self.overrides, PROFILE, profile)
        # Injecting the new profile into the AWS config file retriever.
        aws_config_retriever = self.retrievers.get(AWS_CONFIG_RETRIEVER)
        setattr(aws_config_retriever, PROFILE, profile)

    def set_glue_role_arn(self, glue_role_arn):
        setattr(self.overrides, IAM_ROLE, glue_role_arn)

    def get_profile(self):
        # Attempt to retrieve default profile if a profile is not already set
        profile = self.config_manager.get_config_by_priority(PROFILE)
        if not profile and botocore.session.Session().full_config["profiles"].get("default"):
            setattr(self.overrides, PROFILE, profile)
        return self.config_manager.get_config_by_priority(PROFILE)

    def get_glue_role_arn(self):
        return self.config_manager.get_config_by_priority(IAM_ROLE)

    def get_sessions(self):
        if not self.kernel_gateway.is_glue_client_initialized():
            self.kernel_gateway.initialize_clients(profile=self.get_profile(), region=self.get_region(), endpoint_url=self.get_endpoint_url(), temp_credentials=self.get_temp_credentials())
        return self.kernel_gateway.list_sessions()

    def get_session_id(self):
        return self.session_id

    def get_new_session_id(self):
        return self.new_session_id

    def set_session_id(self, session_id):
        self.session_id = session_id

    def set_new_session_id(self, new_session_id=None):
        self.new_session_id = self._generate_session_id(new_session_id)

    def set_endpoint_url(self, endpoint_url):
        setattr(self.overrides, ENDPOINT_URL, endpoint_url)

    def get_endpoint_url(self):
        return self.config_manager.get_config_by_priority(ENDPOINT_URL)

    def set_region(self, region):
        region = region.lower()
        setattr(self.overrides, REGION, region)

    def get_region(self):
        return self.config_manager.get_config_by_priority(REGION)

    # Stores the pretty name for the SessionType Enum.
    def set_session_type(self, session_type):
        setattr(self.overrides, SESSION_TYPE, session_type)

    def get_session_type(self):
        return self.config_manager.get_config_by_priority(SESSION_TYPE)

    def add_default_argument(self, arg, val):

        if arg in REMAP_DEFAULT_ARGUMENTS:
            arg = REMAP_DEFAULT_ARGUMENTS[arg]

        arg = str(arg) if str(arg).startswith("--") else f'--{arg}'
        val = str(val)

        self.default_arguments[arg] = val

    # Default arguments are not directly exposed to the customer. Hence, config priority is not needed.
    def get_default_arguments(self):
        # Adding additional arguments to the list of default arguments if set by the customer.
        self.add_default_arguments_with_config_priority()

        if self.default_arguments:
            self._send_output(f"Applying the following default arguments:")
            for arg, val in self.default_arguments.items():
                self._send_output(f"{arg} {val}")
        return self.default_arguments

    def add_default_arguments_with_config_priority(self):
        if self.config_manager.get_config_by_priority(ENABLE_DATA_CATALOG):
            self.add_default_argument("enable-glue-datacatalog",self.config_manager.get_config_by_priority(
                ENABLE_DATA_CATALOG))
        if self.config_manager.get_config_by_priority(EXTRA_PY_FILES):
            self.add_default_argument("extra-py-files",self.config_manager.get_config_by_priority(EXTRA_PY_FILES))
        if self.config_manager.get_config_by_priority(EXTRA_JARS):
            self.add_default_argument("extra-jars", self.config_manager.get_config_by_priority(EXTRA_JARS))
        if self.config_manager.get_config_by_priority(ADDITIONAL_PYTHON_MODULES):
            self.add_default_argument("additional-python-modules", self.config_manager.get_config_by_priority(
                ADDITIONAL_PYTHON_MODULES))
        if self.config_manager.get_config_by_priority(SPARK_CONF):
            self.add_default_argument("conf",self.config_manager.get_config_by_priority(SPARK_CONF))

    def set_enable_glue_datacatalog(self):
        setattr(self.overrides, ENABLE_DATA_CATALOG, True)

    def set_extra_py_files(self, extra_py_files):
        setattr(self.overrides, EXTRA_PY_FILES, extra_py_files)

    def set_extra_jars(self, extra_jars):
        setattr(self.overrides, EXTRA_JARS, extra_jars)

    def set_additional_python_modules(self, modules):
        setattr(self.overrides, ADDITIONAL_PYTHON_MODULES, modules)

    def get_connections(self):
        return self.config_manager.get_config_by_priority(CONNECTIONS)

    def set_connections(self, connections):
        setattr(self.overrides, CONNECTIONS, connections)

    def get_account_id(self):
        response = self.kernel_gateway.get_caller_identity(self.get_profile(), self.get_region())
        return response.get('Account')

    def get_tags(self):
        return self.config_manager.get_config_from_all_retrievers(TAGS)

    def create_resource_arn(self):
        region = self.get_region()
        if region in CHINA_REGIONS:
            resource_arn = 'arn:aws-cn:glue:' + str(self.get_region()) + ':' + str(self.get_account_id()) + ':session/' + str(self.get_session_id())
        elif region in US_GOV_REGIONS:
            resource_arn = 'arn:aws-us-gov:glue:' + str(self.get_region()) + ':' + str(self.get_account_id()) + ':session/' + str(self.get_session_id())
        else:
            resource_arn = 'arn:aws:glue:' + str(self.get_region()) + ':' + str(self.get_account_id()) + ':session/' + str(self.get_session_id())
        return resource_arn

    def get_tags_from_resource(self):
        resource_arn = self.create_resource_arn()
        session_id = self.get_session_id()
        try:
            response = self.kernel_gateway.get_tags(resource_arn)
            return response, None
        except botocore.exceptions.ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code in ('AccessDeniedException'):
                return None, "Tags: Unable to display tags due to missing glue:GetTags permission. Please update your IAM policy."
            else:
                return None, f"Unable to fetch tags for the session: {session_id} due to error {error_code}"

    def add_tags_to_resource_ready_state(self, tags):
        resource_arn = self.create_resource_arn()
        session_id = self.get_session_id()
        if tags:
            try:
                response = self.kernel_gateway.tag_resource(resource_arn, tags)
                return None
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('AccessDeniedException'):
                    return "Tags: Unable to add tags due to missing glue:TagResource permission. Please update your IAM policy."
                else:
                    return f"Unable to add tags for the session: {session_id} due to error {error_code}"
        else:
            return

    def set_tags(self, tags):
        if isinstance(tags, str):
            tags_dict = json.loads(tags)
            setattr(self.overrides, TAGS, tags_dict)
        else:
            setattr(self.overrides, TAGS, tags)

    def add_tag(self, key, value):
        tags = getattr(self.overrides, TAGS, defaultdict())
        tags[key] = value

    def add_tag_to_resource(self, tags):
        if tags:
            for (key, value) in tags.items():
                self.add_tag(key, value)

    def get_request_origin(self):
        return self.request_origin

    def set_request_origin(self, request_origin):
        self.request_origin = request_origin

    def get_glue_version(self):
        return self.config_manager.get_config_by_priority(GLUE_VERSION)

    def set_glue_version(self, glue_version):
        setattr(self.overrides, GLUE_VERSION, glue_version)

    def get_session_name(self):
        return self.session_name

    def get_max_capacity(self):
        return self.config_manager.get_config_by_priority(MAX_CAPACITY)

    def set_max_capacity(self, max_capacity):
        setattr(self.overrides, MAX_CAPACITY, max_capacity)

    def get_number_of_workers(self):
        return self.config_manager.get_config_by_priority(NUMBER_OF_WORKERS)

    def set_number_of_workers(self, number_of_workers):
        setattr(self.overrides, NUMBER_OF_WORKERS, number_of_workers)

    def get_worker_type(self):
        return self.config_manager.get_config_by_priority(WORKER_TYPE)

    def set_worker_type(self, worker_type):
        setattr(self.overrides, WORKER_TYPE, worker_type)

    def get_security_config(self):
        return self.config_manager.get_config_by_priority(SECURITY_CONFIG)

    def set_security_config(self, security_config):
        setattr(self.overrides, SECURITY_CONFIG, security_config)

    def get_idle_timeout(self):
        return self.config_manager.get_config_by_priority(IDLE_TIMEOUT)

    def set_idle_timeout(self, idle_timeout):
        setattr(self.overrides, IDLE_TIMEOUT, idle_timeout)

    def set_spark_conf(self, conf):
        setattr(self.overrides, SPARK_CONF, conf)

    def get_spark_conf(self):
        return self.config_manager.get_config_by_priority(SPARK_CONF)

    # TODO: Migrate ray configs using config priority.
    def set_object_memory_head(self, object_memory_head):
        self.add_default_argument("object_store_memory_head", object_memory_head)

    def set_object_memory_worker(self, object_memory_worker):
        self.add_default_argument("object_store_memory_worker", object_memory_worker)

    def set_min_workers(self, min_workers):
        self.add_default_argument("auto-scaling-ray-min-workers", min_workers)

    def set_session_id_prefix(self, session_id_prefix):
        setattr(self.overrides, SESSION_ID_PREFIX, session_id_prefix)

    def get_session_id_prefix(self):
        return self.config_manager.get_config_by_priority(SESSION_ID_PREFIX)

    def _generate_session_id(self, custom_id=None):
        prefix = self.get_session_id_prefix()
        if prefix:
            if not custom_id:
                return f"{prefix}-{uuid.uuid4()}"
            return f"{prefix}-{custom_id}"
        else:
            if not custom_id:
                return f"{uuid.uuid4()}"
            return custom_id

    def create_session(self):

        glue_role_arn = self.get_glue_role_arn()
        if not glue_role_arn:
            glue_arn_from_sts = self.kernel_gateway.get_iam_role_using_sts(self.get_profile(), self.get_region())
            if glue_arn_from_sts:
                self.set_glue_role_arn(glue_arn_from_sts)
            else:
                raise ValueError('Glue Role ARN not provided')

        self._send_output("Trying to create a Glue session for the kernel.")
        if self.get_max_capacity() and (self.get_number_of_workers() and self.get_worker_type()):
            raise ValueError(
                f"Either max_capacity or worker_type and number_of_workers must be set, but not both."
            )

        # Generate new session ID with UUID if no custom ID is set
        if not self.get_new_session_id():
            self.set_new_session_id()
            if not self.get_new_session_id():
                raise ValueError(f"Session ID was not set.")
        session_type_enum = None
        try:
            session_type = self._convert_to_pretty_name(self.get_session_type())
            session_type_enum = SessionType[session_type]
            self._send_output(f"Session Type: {self.get_session_type()}")
        except KeyError:
            raise ValueError(f"Unsupported session type {self.get_session_type()}")

        # If the customer hasn't provided both number of workers and max_capacity, then use the default value.
        if not self.get_max_capacity() and not self.get_number_of_workers() :
            #Both max capacity and number of workers are not provided by the customers.
            #Using the default values from the session type enum.
            self.set_number_of_workers(session_type_enum.default_number_of_workers())

        # If the customer hasn't provided both max_capacity and worker type, then use the default worker type.
        if not self.get_max_capacity() and not self.get_worker_type():
            self.set_worker_type(session_type_enum.default_worker_type())

        # Print Max Capacity if it is set. Else print Number of Workers and Worker Type
        if self.get_max_capacity():
            self._send_output(f"Max Capacity: {self.get_max_capacity()}")
        else:
            self._send_output(f"Worker Type: {self.get_worker_type()}")
            self._send_output(f"Number of Workers: {self.get_number_of_workers()}")

        self._send_output(f"Session ID: {self.get_new_session_id()}")

        additional_args = self._get_additional_arguments()

        response = self.kernel_gateway.create_session(
            self.get_glue_role_arn(),
            self.get_default_arguments(),
            self.get_new_session_id(),
            {"Name": session_type_enum.session_type(), "PythonVersion": session_type_enum.python_version()},
            **additional_args,
        )
        self.set_session_id(
            response["Session"]["Id"]
        )

        self._send_output(
            f"Waiting for session {self.get_session_id()} to get into ready status..."
        )
        start_time = time.time()
        current_status = None
        while time.time() - start_time <= self.time_out:
            current_status = self.get_current_session_status()
            if current_status == READY_SESSION_STATUS:
                break
            elif current_status == PROVISIONING_SESSION_STATUS:
                time.sleep(WAIT_TIME)
                continue
            elif current_status in UNHEALTHY_SESSION_STATUS:
                break
            else:
                time.sleep(WAIT_TIME)
                continue

        if time.time() - start_time > self.time_out:
            self._send_error_output(f"Timed out after {self.time_out} seconds waiting for session to reach {READY_SESSION_STATUS} status.")
        elif current_status in UNHEALTHY_SESSION_STATUS:
            self._send_error_output(f"Session failed to reach {READY_SESSION_STATUS} instead reaching terminal state {self.get_current_session_status()}.")
        elif current_status != READY_SESSION_STATUS:
            self._send_error_output(f"Session failed to reach {READY_SESSION_STATUS} status.")
        else:
            self._send_output(f"Session {self.get_session_id()} has been created.")

    def _get_additional_arguments(self):
        additional_args = {}
        if self.get_max_capacity():
            additional_args["MaxCapacity"] = float(self.get_max_capacity())
        if self.get_number_of_workers():
            additional_args["NumberOfWorkers"] = int(self.get_number_of_workers())
        if self.get_worker_type():
            additional_args["WorkerType"] = self.get_worker_type()
        if self.get_security_config():
            additional_args["SecurityConfiguration"] = self.get_security_config()
        if self.get_idle_timeout():
            additional_args["IdleTimeout"] = int(self.get_idle_timeout())
        if self.get_connections():
            additional_args["Connections"] = {"Connections" : list(self.get_connections().split(","))}
        if self.get_glue_version():
            additional_args["GlueVersion"] = self.get_glue_version()
        if self.get_request_origin():
            additional_args["RequestOrigin"] = self.get_request_origin()

        # add tags
        if self.get_request_origin() == GLUE_STUDIO_NOTEBOOK_IDENTIFIER:
            user_id = self._retrieve_os_env_variable(USER_ID)
        else:
            user_id = self._get_user_id()
        if user_id:  # check if user ID exists and if so, add it as a tag
            self.add_tag(OWNER_TAG, user_id)
        tags = self.get_tags()
        if len(tags) > 0:
            additional_args["Tags"] = tags

        return additional_args

    def stop_session(self):
        if self.get_session_id():
            try:
                self._send_output(f"Stopping session: {self.get_session_id()}")
                # TODO: how do we stop session if our security token expires?
                if not self.kernel_gateway.is_glue_client_initialized():
                    self.kernel_gateway.initialize_clients(profile=self.get_profile(), region=self.get_region(),
                                                           endpoint_url=self.get_endpoint_url(), temp_credentials=self.get_temp_credentials)
                self.kernel_gateway.stop_session(self.get_session_id())
                self.reset_kernel()
            except Exception as e:
                sys.stderr.write(f"Exception encountered while stopping session {self.get_session_id()}: {e} \n")
                self._print_traceback(e)

    def _cancel_statement(self, statement_id: str):
        if not statement_id:
            return

        try:
            self.kernel_gateway.cancel_statement(self.get_session_id(), statement_id)
            start_time = time.time()
            is_ready = False

            while time.time() - start_time <= self.time_out and not is_ready:
                status = self.kernel_gateway.get_statement(self.get_session_id(), statement_id)["Statement"]["State"]
                if status == CANCELLED_STATEMENT_STATUS:
                    self._send_output(f"Statement {statement_id} has been cancelled")
                    is_ready = True

                time.sleep(WAIT_TIME)

            if not is_ready:
                sys.stderr.write(f"Failed to cancel the statement {statement_id}")
        except Exception as e:
            sys.stderr.write(
                f"Exception encountered while canceling statement {statement_id}: {e} \n"
            )
            self._print_traceback(e)

    def get_session_status(self):
        if self.get_session_id() is None:
            return
        else:
            try:
                return self.kernel_gateway.get_session(self.get_session_id())["Session"].get("Status")
            except Exception as e:
                return None

    def get_current_session_status(self):
        try:
            return self.get_current_session()["Status"]
        except Exception as e:
            sys.stderr.write(f"Failed to retrieve session status \n")

    def get_current_session_role(self):
        try:
            return self.get_current_session()["Role"]
        except Exception as e:
            sys.stderr.write(f"Failed to retrieve session role \n")

    def get_current_session(self):
        if self.get_session_id() is None:
            sys.stderr.write(f"No current session.")
        else:
            try:
                current_session = self.kernel_gateway.get_session(self.get_session_id())["Session"]
                return NOT_FOUND_SESSION_STATUS if not current_session else current_session
            except Exception as e:
                sys.stderr.write(f"Exception encountered while retrieving session: {e} \n")
                self._print_traceback(e)

    def _handle_sql_code(self, lines):
        sql_code = "\n".join(lines[1:])
        return f'spark.sql("""{sql_code.rstrip()}""").show()'


    def _send_markdown_output(self, output):
        self.send_response(self.iopub_socket, 'display_data', {
            'metadata': {},
            'data': {
                'text/markdown': output
            }
        })

    def _send_output(self, output):
        stream_content = {"name": "stdout", "text": f"{output}\n"}
        self.send_response(self.iopub_socket, "stream", stream_content)

    def _send_error_output(self, output):
        stream_content = {"name": "stderr", "text": f"{output}\n"}
        self.send_response(self.iopub_socket, "stream", stream_content)

    def _construct_reply_content(self, statement):

        statement_output = statement["Output"]
        status = statement["State"]
        reply_content = {
            "execution_count": self.execution_counter,
            "user_expressions": {},
            "payload": [],
        }

        if status in (AVAILABLE_STATEMENT_STATUS, COMPLETED_STATEMENT_STATUS):
            self.execution_counter += 1
            if status == COMPLETED_STATEMENT_STATUS or ('Status' in statement_output and statement_output["Status"] == "ok"):
                reply_content["status"] = "ok"
                self.display_result(statement_output)
            else:
                reply_content["status"] = "error"
                reply_content.update(
                    {
                        "traceback": statement_output["Traceback"],
                        "ename": statement_output["ErrorName"],
                        "evalue": statement_output["ErrorValue"],
                    }
                )
                self._send_output(
                    f"{statement_output['ErrorName']}: {statement_output['ErrorValue']}"
                )
        elif status == ERROR_STATEMENT_STATUS:
            self.execution_counter += 1
            sys.stderr.write(str(statement_output))
            reply_content.update(self._construct_reply_error(statement_output))
        elif status == FAILED_STATEMENT_STATUS:
            self.execution_counter += 1
            reply_content.update(self._construct_reply_error(statement_output))
            if 'Data' in statement_output:
                statement_output_data = statement_output['Data']
                stdout_text = self.kernel_gateway.get_results(statement_output_data['StdOut']) if 'StdOut' in statement_output_data else ""
                if len(stdout_text):
                    self._send_output(stdout_text)
                failure_reason = self.kernel_gateway.get_results(statement_output_data['Result']) if 'Result' in statement_output_data else ""
                if len(failure_reason):
                    self._send_error_output(failure_reason)
                stderr_text = self.kernel_gateway.get_results(statement_output_data['StdErr']) if 'StdErr' in statement_output_data else ""
                if len(stderr_text):
                    self._send_error_output(stderr_text)

        elif status == CANCELLED_STATEMENT_STATUS:
            self.execution_counter += 1
            reply_content.update(self._construct_reply_error(statement_output))
            self._send_output("This statement is cancelled")

        return reply_content

    def _construct_reply_error(self, statement_output):
        if (not statement_output) or (not all(map(lambda x: x in statement_output, ['TraceBack', 'ErrorName', 'ErrorValue']))):
            return dict({"status": "abort"})
        else:
            return dict({
                "status": "error",
                "traceback": statement_output["Traceback"],
                "ename": statement_output["ErrorName"],
                "evalue": statement_output["ErrorValue"],
            })

    async def _do_execute(self, code, silent, store_history, user_expressions, allow_stdin):
        res = await self._execute_cell(code, silent, store_history, user_expressions, allow_stdin)
        return res

    async def _execute_cell(
            self, code, silent, store_history=True, user_expressions=None, allow_stdin=False
    ):
        reply_content = await self._execute_cell_for_user(
            code, silent, store_history, user_expressions, allow_stdin
        )
        return reply_content

    async def _execute_cell_for_user(
            self, code, silent, store_history=True, user_expressions=None, allow_stdin=False
    ):
        result = await super(BaseKernel, self).do_execute(
            code, silent, store_history, user_expressions, allow_stdin
        )
        if isinstance(result, Future):
            result = result.result()
        return result

    async def _execute_magics(self, code, silent, store_history, user_expressions, allow_stdin):
        try:
            magic_lines = 0
            lines = code.splitlines()
            for line in lines:
                # If there is a cell magic, we simply treat all the remaining code as part of the cell magic
                if any(line.startswith(cell_magic) for cell_magic in CELL_MAGICS):
                    if line.startswith(SQL_CELL_MAGIC):
                        return self._handle_sql_code(lines)
                    else:
                        code = "\n".join(lines[magic_lines:])
                        await self._do_execute(
                            code, silent, store_history, user_expressions, allow_stdin
                        )
                    return None
                # If we encounter a line magic, we execute this line magic and continue
                if line.startswith("%") or line.startswith("!"):
                    await self._do_execute(
                        line, silent, store_history, user_expressions, allow_stdin
                    )
                    magic_lines += 1
                # We ignore comments and empty lines
                elif line.startswith("#") or not line:
                    magic_lines += 1
                else:
                    break
            code = "\n".join(lines[magic_lines:])
            return code
        except Exception as e:
            sys.stderr.write(f"Exception encountered: {e} \n")
            self._print_traceback(e)
            return await self._complete_cell()

    async def _complete_cell(self):
        """A method that runs a cell with no effect. Call this and return the value it
        returns when there's some sort of error preventing the user's cell from executing; this
        will register the cell from the Jupyter UI as being completed."""
        return await self._execute_cell("None", False, True, None, False)

    def _register_magics(self):
        ip = get_ipython()
        magics = KernelMagics(ip, "", self)
        ip.register_magics(magics)

    def _initialize_retrievers(self, overrides):
        # Config priority is Overridden Config (Magics) > Env variables > Config File > Defaults.
        # OrderedDict is explicitly used to show config priority.
        retrievers = OrderedDict()
        retrievers[OVERRIDES_RETRIEVER] = OverriddenConfigRetriever(overrides)
        retrievers[SYSTEM_ENV_RETRIEVER] = SystemEnvConfigRetriever(self.get_request_origin())
        retrievers[AWS_CONFIG_RETRIEVER] = AWSFileConfigRetriever()
        retrievers[DEFAULT_RETRIEVER] = DefaultConfigRetriever(default_configs)
        return retrievers

    def _print_traceback(self, e):
        traceback.print_exception(type(e), e, e.__traceback__)

    async def _print_startup_text(self):
        self._send_output("Welcome to the Glue Interactive Sessions Kernel")
        self._send_output(
            "For more information on available magic commands, please type %help in any new cell.\n"
        )
        self._send_output(
            "Please view our Getting Started page to access the most up-to-date information on the Interactive Sessions kernel: https://docs.aws.amazon.com/glue/latest/dg/interactive-sessions.html"
        )
        self.should_print_startup_text = False
        current_kernel_version = self._get_current_kernel_version("aws-glue-sessions")
        latest_kernel_version = self._get_latest_kernel_version("aws-glue-sessions")
        if (
                latest_kernel_version
                and current_kernel_version
                and latest_kernel_version != current_kernel_version
        ):
            self._send_output(
                f"It looks like there is a newer version of the kernel available. The latest version is {latest_kernel_version} and you have {current_kernel_version} installed."
            )
            self._send_output(
                "Please run `pip install --upgrade aws-glue-sessions` to upgrade your kernel"
            )
        elif latest_kernel_version is None and current_kernel_version is not None:
            self._send_output(f"Installed kernel version: {current_kernel_version} ")

    def _get_latest_kernel_version(self, name):
        if not (self.request_origin == "GlueStudioNotebook"
                or self.request_origin in SAGEMAKER_STUDIO_NOTEBOOK_IDENTIFIERS):
            try:
                response = requests.get("https://pypi.org/pypi/{}/json".format(name))
                data = response.json()
                latest_version = data["info"]["version"]
                return str(latest_version)
            except Exception:
                return None

    def _get_current_kernel_version(self, name):
        try:
            current_version = version(name)
            return str(current_version)
        except Exception:
            return None

    def reset_kernel(self):
        self.set_session_id(None)
        self.set_new_session_id(None)
        self.execution_counter = 1
        self.kernel_gateway = KernelGateway(self.ipython_display)

    # https://github.com/ipython/ipykernel/issues/795
    # Redirect logs to only print in terminal

    def _setup_terminal_logging(self):
        for std, __std__ in [
            (sys.stdout, sys.__stdout__),
            (sys.stderr, sys.__stderr__),
        ]:
            if getattr(std, "_original_stdstream_copy", None) is not None:
                # redirect captured pipe back to original FD
                os.dup2(std._original_stdstream_copy, __std__.fileno())
                std._original_stdstream_copy = None

    def _log_to_terminal(self, log):
        print(f"LOG: {log}", file=sys.__stdout__, flush=True)

    def _get_user_id(self):
        response = self.kernel_gateway.get_caller_identity(self.get_profile(), self.get_region())
        return response.get("UserId")

    def _retrieve_from_aws_config(self, key):
        custom_profile_session = botocore.session.Session(profile=self.get_profile())
        return custom_profile_session.full_config["profiles"][self.get_profile()].get(key)

    def _retrieve_os_env_variable(self, key):
        if self.get_request_origin() not in SAGEMAKER_STUDIO_NOTEBOOK_IDENTIFIERS:
            return os.environ.get(key)
        else:
            _, output = subprocess.getstatusoutput(f"echo ${key}")
            return os.getenv(key, output)

    def handle_exception(self, e):
        print(Colour.Red + "Error: " + str(e))

    def _format_endpoint_url(self, region):
        if region in CHINA_REGIONS:
            return f"https://glue.{region}.amazonaws.com.cn"
        return f"https://glue.{region}.amazonaws.com"

    def _convert_to_pretty_name(self, session_type):
        # Some systems are already passing the session_type as the actual value and other systems are passing it as pretty name.
        # Handling both the cases irrespective of what's given, converting it to pretty name and looking up the ENUM
        if session_type.lower() in [SessionType.etl.session_type().lower(), SessionType.etl.pretty_name().lower()]:
            return SessionType.etl.pretty_name()
        elif session_type.lower() in [SessionType.glue_ray.session_type().lower(), SessionType.glue_ray.pretty_name().lower()]:
            return SessionType.glue_ray.pretty_name()
        elif session_type.lower() in [SessionType.streaming.session_type().lower(), SessionType.streaming.pretty_name().lower()]:
            return SessionType.streaming.pretty_name()




import importlib
from importlib import abc
import sys
import os
import functools
import pathlib

def patchonimport(module_name=None):
    def decorator(target_function):
        sys.meta_path.insert(0, ImportModulePatch({module_name: target_function}))

        def wrapper(*args, **kwargs):
            target_function(*args, **kwargs)

        return wrapper

    return decorator


class ImportModulePatch(abc.MetaPathFinder, abc.Loader):
    def __init__(self, patch_modules):
        # type: (Dict[str, Callable[[Module], None]]) -> None
        self._patch_modules = patch_modules
        self._in_create_module = False

    def find_module(self, fullname, path=None):
        spec = self.find_spec(fullname, path)
        if spec is None:
            return None
        return spec

    def create_module(self, spec):
        try:
            self._in_create_module = True

            from importlib.util import find_spec, module_from_spec
            real_spec = importlib.util.find_spec(spec.name)

            real_module = module_from_spec(real_spec)
            real_spec.loader.exec_module(real_module)

            self._in_create_module = False
            return real_module
        except:
            return None

    def exec_module(self, module):
        try:
            _ = sys.modules.pop(module.__name__)
        except KeyError:
            print("module %s is not in sys.modules", module.__name__)
        sys.modules[module.__name__] = module
        globals()[module.__name__] = module
        self._patch_modules[module.__name__](module)

    def find_spec(self, fullname, path=None, target=None):
        if fullname not in self._patch_modules:
            return None
        if self._in_create_module:
            return None
        spec = importlib.machinery.ModuleSpec(fullname, self)
        return spec


class PatchesForRay(object):
    @staticmethod
    @patchonimport(module_name='botocore')
    def loader_patch(real_module):
        os.environ['AWS_DATA_PATH'] = str(pathlib.Path(__file__).parent)

# Importing base kernel after the patch so that botocore is patched on import
from ..glue_kernel_base.BaseKernel import BaseKernel


class GlueKernel(BaseKernel):
    implementation = "Python Glue Session"
    implementation_version = "1.0"
    language = "no-op"
    language_version = "0.1"
    language_info = {
        "name": "Python_Glue_Session",
        "mimetype": "text/x-python",
        "codemirror_mode": {"name": "python", "version": 3},
        "pygments_lexer": "python3",
        "file_extension": ".py",
    }
    session_language = "python"

    def __init__(self, **kwargs):
        self.request_origin = "GluePySparkKernel"
        super(GlueKernel, self).__init__(**kwargs)


if __name__ == "__main__":
    from ipykernel.kernelapp import IPKernelApp

    IPKernelApp.launch_instance(kernel_class=GlueKernel)



import json

from aws_glue_interactive_sessions_kernel.glue_kernel_configs.ConfigRetriever import ConfigRetriever
from aws_glue_interactive_sessions_kernel.glue_kernel_utils.GlueSessionsConstants import \
    SAGEMAKER_STUDIO_NOTEBOOK_IDENTIFIERS, GLUE_PREFIX, GLUE_VERSION, REGION, SESSION_ID_PREFIX, SPARK_CONF, TAGS, \
    SESSION_TYPE, GLUE_JOB_TYPE, IAM_ROLE, GLUE_ROLE_ARN, EXTRA_PY_FILES, ADDITIONAL_PYTHON_MODULES, EXTRA_JARS, CONNECTIONS
from aws_glue_interactive_sessions_kernel.glue_kernel_utils.ValidationUtils import convert_list_to_string

import os
import subprocess


class SystemEnvConfigRetriever(ConfigRetriever):

    def __init__(self, request_origin=None):
        self.request_origin = request_origin

    def _retrieve_os_env_variable(self, key):
        if self.request_origin not in SAGEMAKER_STUDIO_NOTEBOOK_IDENTIFIERS:
            return os.environ.get(key)
        else:
            _, output = subprocess.getstatusoutput(f"echo ${key}")
            value = os.getenv(key, output)
            return value if value else None

    def get_config_variable(self, config_name):
        # Special handling for these configs which are already used.
        # Accept both the old and new values to have a gradual migration.
        if config_name in [GLUE_VERSION, REGION, SPARK_CONF, SESSION_ID_PREFIX]:
            return self._retrieve_os_env_variable(config_name) \
                if self._retrieve_os_env_variable(config_name) is not None \
                else self._retrieve_os_env_variable(GLUE_PREFIX + config_name)
        # Special handling for session type as the config is already defined as a job_type environment variable.
        if config_name == SESSION_TYPE:
            return self._retrieve_os_env_variable(GLUE_JOB_TYPE) \
                if self._retrieve_os_env_variable(GLUE_JOB_TYPE) \
                else self._retrieve_os_env_variable(GLUE_PREFIX + SESSION_TYPE)
        if config_name == IAM_ROLE:
            return self._retrieve_os_env_variable(GLUE_ROLE_ARN)  \
                if self._retrieve_os_env_variable(GLUE_ROLE_ARN) is not None \
                else self._retrieve_os_env_variable(GLUE_PREFIX + IAM_ROLE)
        if config_name == TAGS:
            return self._retrieve_tags_from_env_variable(GLUE_PREFIX + TAGS)

        if config_name in [EXTRA_PY_FILES, ADDITIONAL_PYTHON_MODULES, EXTRA_JARS, CONNECTIONS]:
            return convert_list_to_string(self._retrieve_os_env_variable(GLUE_PREFIX + config_name)) \
                if self._retrieve_os_env_variable(GLUE_PREFIX + config_name) is not None \
                else None
        return self._retrieve_os_env_variable(GLUE_PREFIX + config_name)

    def _retrieve_tags_from_env_variable(self, tags_key):
        tags = self._retrieve_os_env_variable(tags_key)
        if isinstance(tags, str):
            tags_dict = json.loads(tags)
            return tags_dict
        else:
            return tags




import os
import pathlib

import click

from jupyter_client.kernelspec import install_kernel_spec
from rich.console import Console
from rich.markdown import Markdown


ERROR_MESSAGE = """
# Automatic installation of jupyter kernel spec failed

To overcome this error, please install the kernelspec manually.

### MacOS/Linux instructions
Please run the following command on the terminal 

```
SITE_PACKAGES=$(pip3 show aws-glue-sessions | grep Location | awk '{print $2}')
  jupyter kernelspec install $SITE_PACKAGES/aws_glue_interactive_sessions_kernel/glue_pyspark
  jupyter kernelspec install $SITE_PACKAGES/aws_glue_interactive_sessions_kernel/glue_spark 
```

### Windows Instructions

1. Change the directory to the aws-glue-sessions install directory within python's site-packages directory.
    Windows PowerShell:
    ```
    cd ((pip3 show aws-glue-sessions | Select-String Location | % {$_ -replace("Location: ","")})+"\aws_glue_interactive_sessions_kernel")
    ```
2. Install the AWS Glue PySpark and AWS Glue Scala kernels.
    ```
    jupyter-kernelspec install glue_pyspark
    ```
    ```
    jupyter-kernelspec install glue_spark
    ```
"""


@click.group()
def cli():
    pass

@click.command()
def install():
    do_install()
def do_install_kernel_spec(kernel_name):
    print(f'Installing jupyter kernelspec for {kernel_name}')
    curr_path = pathlib.Path(__file__).parent.resolve()
    # Documentation : https://jupyter-client.readthedocs.io/en/latest/kernels.html#kernelspecs
    install_kernel_spec(str(curr_path.joinpath(f'{kernel_name}')), user=True, replace=True)
    print(f'Installed kernelspec for {kernel_name}')

def do_install():
    try:
        do_install_kernel_spec('glue_pyspark')
        do_install_kernel_spec('glue_spark')
    except Exception as ex:
        console = Console()
        md = Markdown(ERROR_MESSAGE)
        console.print(md)



if __name__ == '__main__':
    install()



from aws_glue_interactive_sessions_kernel.glue_kernel_utils.GlueSessionsConstants import SESSION_TYPE
from ..glue_kernel_base.BaseKernel import BaseKernel
from ..glue_kernel_utils.GlueSessionsConstants import SessionType


class GlueKernel(BaseKernel):
    implementation = "Scala Glue Session"
    implementation_version = "1.0"
    language = "no-op"
    language_version = "0.1"
    language_info = {
        "name": "scala",
        "mimetype": "text/x-scala",
        "codemirror_mode": "text/x-scala",
        "pygments_lexer": "scala",
    }
    session_language = "scala"

    def __init__(self, **kwargs):
        self.request_origin = "GlueSparkKernel"
        super(GlueKernel, self).__init__(**kwargs)

    def _set_default_arguments(self):
        self.default_arguments = {
            "--glue_kernel_version": self._get_current_kernel_version("aws-glue-sessions"),
            "--session-language": "scala",
            "--enable-glue-datacatalog": "true",
        }

    def set_python_version(self, python_version):
        pass

    def set_object_memory_head(self, object_memory_head):
        pass

    def set_object_memory_worker(self, object_memory_worker):
        pass

    def set_min_workers(self, min_workers):
        pass

    def set_session_type(self, session_type):
        if session_type in [SessionType.glue_ray.pretty_name(), SessionType.glue_ray.session_type()]:
            self._send_error_output(
                f'The requested session type "{session_type}" is not supported by Scala kernel.')
            return
        setattr(self.overrides, SESSION_TYPE, session_type)

if __name__ == "__main__":
    from ipykernel.kernelapp import IPKernelApp

    IPKernelApp.launch_instance(kernel_class=GlueKernel)



import json

from aws_glue_interactive_sessions_kernel.glue_kernel_configs.ConfigRetriever import ConfigRetriever

import botocore

from aws_glue_interactive_sessions_kernel.glue_kernel_utils.GlueSessionsConstants import GLUE_PREFIX, GLUE_VERSION, \
    REGION, SESSION_TYPE, SPARK_CONF, SESSION_ID_PREFIX, TAGS, GLUE_JOB_TYPE, IAM_ROLE, GLUE_ROLE_ARN, \
    EXTRA_PY_FILES, EXTRA_JARS, CONNECTIONS, ADDITIONAL_PYTHON_MODULES
from aws_glue_interactive_sessions_kernel.glue_kernel_utils.ValidationUtils import convert_list_to_string



class AWSFileConfigRetriever(ConfigRetriever):
    def __init__(self, profile=None):
        self.profile = profile

    def get_profile(self):
        # Attempt to retrieve default profile if a profile is not already set
        if not self.profile and botocore.session.Session().full_config["profiles"].get("default"):
            self.profile = "default"
        return self.profile

    def _retrieve_from_aws_config(self, key):
        profile = self.get_profile()
        if profile:
            custom_profile_session = botocore.session.Session(profile=self.get_profile())
            if custom_profile_session.full_config["profiles"].get(profile):
                return custom_profile_session.full_config["profiles"][self.get_profile()].get(key)
        return None
    def get_config_variable(self, config_name):
        # Special handling for these configs where the configs are already used by customers and cannot be changed
        if config_name in [GLUE_VERSION, REGION, SPARK_CONF, SESSION_ID_PREFIX]:
            return self._retrieve_from_aws_config(config_name) \
                if self._retrieve_from_aws_config(config_name) is not None \
                else self._retrieve_from_aws_config(GLUE_PREFIX + config_name)
        # Special handling for session type as the config is already defined as a job_type environment variable.
        if config_name == SESSION_TYPE:
            return self._retrieve_from_aws_config(GLUE_JOB_TYPE) \
                if self._retrieve_from_aws_config(GLUE_JOB_TYPE) \
                else self._retrieve_from_aws_config(GLUE_PREFIX + SESSION_TYPE)

        if config_name == IAM_ROLE:
            return self._retrieve_from_aws_config(GLUE_ROLE_ARN) \
                if self._retrieve_from_aws_config(GLUE_ROLE_ARN) is not None \
                else self._retrieve_from_aws_config(GLUE_PREFIX + IAM_ROLE)
        if config_name == TAGS:
            return self._retrieve_tags_from_aws_config(GLUE_PREFIX + TAGS)

        if config_name in [EXTRA_PY_FILES, ADDITIONAL_PYTHON_MODULES, EXTRA_JARS, CONNECTIONS]:
            return convert_list_to_string(self._retrieve_from_aws_config(GLUE_PREFIX + config_name)) \
                if self._retrieve_from_aws_config(GLUE_PREFIX + config_name) is not None \
                else None
        return self._retrieve_from_aws_config(GLUE_PREFIX + config_name)

    def _retrieve_tags_from_aws_config(self, tags_key):
        tags = self._retrieve_from_aws_config(tags_key)
        if isinstance(tags, str):
            tags_dict = json.loads(tags)
            return tags_dict
        else:
            return tags




from botocore.config import Config

import time
import traceback

import re

import boto3
import botocore
from botocore.exceptions import ClientError
from urllib3.util import parse_url

from ..glue_kernel_utils.GlueSessionsConstants import *


def split_s3_path(s3_path):
    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)
    return bucket, key


class KernelGateway:
    # Configuring max_ties to 3 and min retry between calls to 0.5 seconds for retryable exceptions.
    max_tries = 3
    min_retry_time = 0.5

    def __init__(self, display):
        self.display = display
        self.glue_client = None
        self.sts_client = None
        self.iam_client = None
        self.s3_client = None

    def initialize_clients(self, profile=None, region=None, endpoint_url=None, temp_credentials=None):
        if temp_credentials:
            self.initialize_clients_with_temp_credentials(temp_credentials, region)
            return
        self.sts_client = self.create_sts_client(profile, region)
        self.iam_client = self.create_iam_client(profile, region)
        self.s3_client = self.create_s3_client(profile, region)
        self.glue_client = self._authenticate(profile, region, endpoint_url)

    def initialize_clients_with_temp_credentials(self, temp_credentials, region):
        self.sts_client = boto3.Session().client("sts", region_name = region,
                                                 aws_access_key_id=temp_credentials['temp_aws_access_key_id'] ,
                                                 aws_secret_access_key=temp_credentials['temp_aws_secret_access_key'],
                                                 aws_session_token=temp_credentials['temp_aws_session_token'])
        self.iam_client = boto3.Session().client("iam", region_name = region,
                                                 aws_access_key_id=temp_credentials['temp_aws_access_key_id'] ,
                                                 aws_secret_access_key=temp_credentials['temp_aws_secret_access_key'],
                                                 aws_session_token=temp_credentials['temp_aws_session_token'])
        self.s3_client = boto3.Session().client("s3", region_name = region,
                                               aws_access_key_id=temp_credentials['temp_aws_access_key_id'] ,
                                               aws_secret_access_key=temp_credentials['temp_aws_secret_access_key'],
                                               aws_session_token=temp_credentials['temp_aws_session_token'])
        self.glue_client = boto3.Session().client("glue", region_name = region,
                                                  aws_access_key_id=temp_credentials['temp_aws_access_key_id'] ,
                                                  aws_secret_access_key=temp_credentials['temp_aws_secret_access_key'],
                                                  aws_session_token=temp_credentials['temp_aws_session_token'])


    def _authenticate(self, profile=None, region=None, endpoint_url=None):
        # region must be set
        if not region:
            raise ValueError(f"Region must be set.")
        # If we are using a custom endpoint
        if not endpoint_url:
            endpoint_url = self._format_endpoint_url(region)
        if profile:
            return self._authenticate_with_profile(profile, region, endpoint_url)
        else:
            return self._create_client(profile, region, endpoint_url)

    def _authenticate_with_profile(self, profile=None, region=None, endpoint_url=None):
        if profile not in self._get_available_profiles():
            raise ValueError(f"Profile {profile} not defined in config")
        return self._create_client(profile, region, endpoint_url)

    def _create_client(self, profile=None, region=None, endpoint_url=None):
        # Having a standard retry config for the boto3 client.
        # Retries on the following AWS Standard errors/exceptions:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html#standard-retry-mode
        config = Config(
            retries={
                'max_attempts': 1,
                'mode': 'standard'
            }
        )
        # boto3 will automatically look for env variables, config file, IMDS
        # for credentials in the order: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
        if profile:
            return boto3.Session(profile_name=profile).client("glue", region_name=region, config=config)
        else:
            return boto3.Session().client("glue", region_name=region, endpoint_url=endpoint_url, config=config)

    def create_session(self, role, default_arguments, new_session_id, command, **additional_args):
        for i in range(1, self.max_tries + 1):
            try:
                response = self.glue_client.create_session(
                    Role=role,
                    DefaultArguments=default_arguments,
                    Id=new_session_id,
                    Command=command,
                    **additional_args,
                )
                return response
                # All boto3 exceptions are classified under ClientError.
                # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/error-handling.html
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                error_message = e.response['Error']['Message']
                if error_code in ('OperationTimeoutException', 'InternalServiceException',
                                  'ResourceNumberLimitExceededException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                elif error_code in ('InvalidInputException', 'AlreadyExistsException', 'AccessDeniedException'):
                    self.display.send_error(f'Following exception encountered while creating session: {e} \n')
                    self.display.send_error(f'Error message: {error_message} \n')
                    self._print_traceback(e)
                    raise
                else:
                    self.display.send_error(f'Unknown exception encountered while creating session: {e} \n')
                    self.display.send_error(f'Error message: {error_message} \n')
                    self._print_traceback(e)
                    raise
        raise RuntimeError('Error while creating session')

    def stop_session(self, session_id):
        for i in range(1, self.max_tries + 1):
            if session_id:
                try:
                    response = self.glue_client.stop_session(Id=session_id)
                    return response
                except botocore.exceptions.ClientError as e:
                    error_code = e.response['Error']['Code']
                    if error_code in ('OperationTimeoutException', 'InternalServiceException'):
                        time.sleep(i * self.min_retry_time)
                        continue
                    else:
                        raise
        raise RuntimeError('Error while stopping session')

    def run_statement(self, session_id, code):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.run_statement(
                    SessionId=session_id, Code=code
                )
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'ResourceNumberLimitExceededException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while running statement')

    def cancel_statement(self, session_id, statement_id: str):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.cancel_statement(SessionId=session_id, Id=statement_id)
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'OperationTimeoutException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while canceling statement')

    def get_statement(self, session_id, statement_id: str):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.get_statement(SessionId=session_id, Id=statement_id)
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'OperationTimeoutException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while getting statement')

    def get_session(self, session_id):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.get_session(Id=session_id)
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'OperationTimeoutException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while getting session')

    def list_sessions(self):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.list_sessions()
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'OperationTimeoutException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while getting list of session')

    def get_tags(self, resource_arn):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.get_tags(ResourceArn=resource_arn).get("Tags")
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'OperationTimeoutException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while fetching tags for the session')

    def tag_resource(self, resource_arn, tags):
        for i in range(1, self.max_tries + 1):
            try:
                return self.glue_client.tag_resource(ResourceArn=resource_arn, TagsToAdd=tags)
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code in ('InternalServiceException', 'OperationTimeoutException'):
                    time.sleep(i * self.min_retry_time)
                    continue
                else:
                    raise
        raise RuntimeError('Error while adding tags for the session')

    def get_results(self, s3_path):
        if not s3_path.startswith("s3"):
            print(f"No output file found at {s3_path}")
            return "No output file found"
        try:
            urlparts = parse_url(s3_path)
            s3_object = self.s3_client.get_object(Bucket=urlparts.host, Key=urlparts.path[1:])
            return s3_object["Body"].read().decode("utf-8")
        except botocore.exceptions.ClientError as e:
            print(f"Response from get_object with {s3_path} was {e}")
            raise RuntimeError("Error while getting calculation results", e)


    def _get_available_profiles(self):
        return boto3.session.Session().available_profiles

    def is_glue_client_initialized(self):
        if self.glue_client is None:
            return False
        return True

    def _print_traceback(self, e):
        traceback.print_exception(type(e), e, e.__traceback__)

    def get_iam_role_using_sts(self, profile=None, region=None):
        try:
            role_arn = self.sts_client.get_caller_identity().get("Arn")
        except Exception:
            return None
        regex = r"arn:aws[^:]*:sts::[0-9]*:assumed-role/(.+)/.+"
        m = re.match(regex, role_arn)
        if m:
            role_arn = self._get_role_arn_from_iam(m.group(1), region)
            return role_arn

        return None

    def _get_role_arn_from_iam(self, role_name, region=None):
        return self.iam_client.get_role(RoleName=role_name).get("Role", {}).get("Arn", None)

    def get_caller_identity(self, profile=None, region=None):
        return self.sts_client.get_caller_identity()

    def create_sts_client(self, profile=None, region=None):
        if not region:
            raise ValueError(f"Region must be set.")
        if profile:
            return boto3.Session(profile_name=profile).client(
                "sts", region_name=region
            )
        else:
            return boto3.Session().client(
                "sts",
                region_name=region,
                endpoint_url=self._get_sts_endpoint_url(region),
            )

    def create_iam_client(self, profile=None, region=None):
        if not region:
                raise ValueError(f"Region must be set.")
        if profile:
            return boto3.Session(profile_name=profile).client(
                "iam", region_name=region
            )
        else:
            return boto3.Session().client(
                "iam",
                region_name=region
            )

    def create_s3_client(self, profile=None, region=None):
        if not region:
                raise ValueError(f"Region must be set.")
        if profile:
            return boto3.Session(profile_name=profile).client(
                "s3", region_name=region,
                config=Config(retries=dict(max_attempts=10))
            )
        else:
            return boto3.Session().client(
                "s3",
                region_name=region,
                config=Config(retries=dict(max_attempts=10))
            )

    def _get_sts_endpoint_url(self, region):
        if region in CHINA_REGIONS:
            return f"https://sts.{region}.amazonaws.com.cn"
        return f"https://sts.{region}.amazonaws.com"

    def _format_endpoint_url(self, region):
        if region in CHINA_REGIONS:
            return f"https://glue.{region}.amazonaws.com.cn"
        return f"https://glue.{region}.amazonaws.com"





import pathlib
from enum import Enum, unique
from tabulate import tabulate


class SessionTable:
    def __init__(self):
        self.table_data = []

    def add_session_info(self, session_id, status, created_on):
        self.table_data.append([session_id, status, created_on])

    def format_session_info(self):
        table_headers = ["SessionId", "Status", "CreatedOn"]
        table = tabulate(self.table_data, headers=table_headers, tablefmt="grid")
        return table


# Symbols for referencing Mime types
class MimeTypes(Enum):
    TextPlain = ("text/plain")
    ImagePng = ("image/png")
    S3URI = ("text/uri-list")

# Symbols for referencing session types
@unique
class SessionType(Enum):

    def __new__(cls, value, pretty_name, python_version , default_number_of_workers, default_worker_type):
        obj = object.__new__(cls)
        obj._value_ = value
        obj._pretty_name = pretty_name
        obj._python_version = python_version
        obj._default_number_of_workers = default_number_of_workers
        obj._default_worker_type = default_worker_type
        return obj

    etl = ("glueetl", 'etl',  "3", "5", "G.1X")
    streaming = ("gluestreaming", 'streaming', "3", "5", "G.1X")
    glue_ray = ("glueray", 'glue_ray', "3.9", "5", "Z.2X")

    def python_version(self):
        return self._python_version

    def session_type(self):
        return self._value_

    def pretty_name(self):
        return self._pretty_name
    def default_number_of_workers(self):
        return self._default_number_of_workers

    def default_worker_type(self):
        return self._default_worker_type


WAIT_TIME = 1

READY_SESSION_STATUS = "READY"
PROVISIONING_SESSION_STATUS = "PROVISIONING"
NOT_FOUND_SESSION_STATUS = "NOT_FOUND"
FAILED_SESSION_STATUS = "FAILED"
STOPPING_SESSION_STATUS = "STOPPING"
STOPPED_SESSION_STATUS = "STOPPED"
TIMEOUT_SESSION_STATUS = "TIMEOUT"
UNHEALTHY_SESSION_STATUS = [NOT_FOUND_SESSION_STATUS, FAILED_SESSION_STATUS, STOPPING_SESSION_STATUS,
                            STOPPED_SESSION_STATUS]

ERROR_STATEMENT_STATUS = "ERROR"
FAILED_STATEMENT_STATUS = "FAILED"
CANCELLED_STATEMENT_STATUS = "CANCELLED"
AVAILABLE_STATEMENT_STATUS = "AVAILABLE"
COMPLETED_STATEMENT_STATUS = "COMPLETED"
FINAL_STATEMENT_STATUS = [FAILED_STATEMENT_STATUS, ERROR_STATEMENT_STATUS, CANCELLED_STATEMENT_STATUS,
                          AVAILABLE_STATEMENT_STATUS, COMPLETED_STATEMENT_STATUS]
SQL_CELL_MAGIC = "%%sql"

CELL_MAGICS = {"%%configure", "%%sql", "%%ray_conf", "%%tags", "%%assume_role"}

VALID_GLUE_VERSIONS = {"2.0", "3.0", "4.0"}

CHINA_REGIONS = {"cn-north-1", "cn-northwest-1"}

US_GOV_REGIONS = {"us-gov-east-1", "us-gov-west-1"}


class Colour:
    Red = "\033[31m"
    Green = "\033[32m"
    Yellow = "\033[33m"
    Blue = "\033[34m"
    Magenta = "\033[35m"
    Cyan = "\033[36m"
    LightGray = "\033[37m"
    DarkGray = "\033[90m"
    LightRed = "\033[91m"
    LightGreen = "\033[92m"
    LightYellow = "\033[93m"
    LightBlue = "\033[94m"
    LightMagenta = "\033[95m"
    LightCyan = "\033[96m"
    White = "\033[97m"
    Bold = "\033[1m"
    Reset = "\033[0m"


HELP_TEXT = f'''
# Available Magic Commands

## Sessions Magic

----
    %help                             Return a list of descriptions and input types for all magic commands. 
    %profile            String        Specify a profile in your aws configuration to use as the credentials provider.
    %region             String        Specify the AWS region in which to initialize a session. 
                                      Default from ~/.aws/config on Linux or macOS, 
                                      or C:\\Users\\ USERNAME \\.aws\\config" on Windows.
    %idle_timeout       Int           The number of minutes of inactivity after which a session will timeout. 
                                      Default: 2880 minutes (48 hours).
    %session_id_prefix  String        Define a String that will precede all session IDs in the format 
                                      [session_id_prefix]-[session_id]. If a session ID is not provided,
                                      a random UUID will be generated.
    %status                           Returns the status of the current Glue session including its duration, 
                                      configuration and executing user / role.
    %session_id                       Returns the session ID for the running session. 
    %list_sessions                    Lists all currently running sessions by ID.
    %stop_session                     Stops the current session.
    %glue_version       String        The version of Glue to be used by this session. 
                                      Currently, the only valid options are 2.0, 3.0 and 4.0. 
                                      Default: 2.0.
----

## Selecting Session Types

----
    %streaming          String        Sets the session type to Glue Streaming.
    %etl                String        Sets the session type to Glue ETL.
    %glue_ray           String        Sets the session type to Glue Ray.
    %session_type       String        Specify a session_type to be used. Supported values: streaming, etl and glue_ray. 
----

## Glue Config Magic 
*(common across all session types)*

----

    %%configure         Dictionary    A json-formatted dictionary consisting of all configuration parameters for 
                                      a session. Each parameter can be specified here or through individual magics.
    %iam_role           String        Specify an IAM role ARN to execute your session with.
                                      Default from ~/.aws/config on Linux or macOS, 
                                      or C:\\Users\\%USERNAME%\\.aws\\config` on Windows.
    %number_of_workers  int           The number of workers of a defined worker_type that are allocated 
                                      when a session runs.
                                      Default: 5.
    %additional_python_modules  List  Comma separated list of additional Python modules to include in your cluster 
                                      (can be from Pypi or S3).
    %%tags        Dictionary          Specify a json-formatted dictionary consisting of tags to use in the session.
    
    %%assume_role Dictionary, String  Specify a json-formatted dictionary or an IAM role ARN string to create a session 
                                      for cross account access.
                                      E.g. {{valid arn}}
                                      %%assume_role 
                                      'arn:aws:iam::XXXXXXXXXXXX:role/AWSGlueServiceRole' 
                                      E.g. {{credentials}}
                                      %%assume_role
                                      {{
                                            "aws_access_key_id" : "XXXXXXXXXXXX",
                                            "aws_secret_access_key" : "XXXXXXXXXXXX",
                                            "aws_session_token" : "XXXXXXXXXXXX"
                                       }}
----

                                      
## Magic for Spark Sessions (ETL & Streaming)

----
    %worker_type        String        Set the type of instances the session will use as workers. 
    %connections        List          Specify a comma separated list of connections to use in the session.
    %extra_py_files     List          Comma separated list of additional Python files From S3.
    %extra_jars         List          Comma separated list of additional Jars to include in the cluster.
    %spark_conf         String        Specify custom spark configurations for your session. 
                                      E.g. %spark_conf spark.serializer=org.apache.spark.serializer.KryoSerializer
----
                                      
## Magic for Ray Session

----
    %min_workers        Int           The minimum number of workers that are allocated to a Ray session. 
                                      Default: 1.
    %object_memory_head Int           The percentage of free memory on the instance head node after a warm start. 
                                      Minimum: 0. Maximum: 100.
    %object_memory_worker Int         The percentage of free memory on the instance worker nodes after a warm start. 
                                      Minimum: 0. Maximum: 100.
----

## Action Magic

----

    %%sql               String        Run SQL code. All lines after the initial %%sql magic will be passed
                                      as part of the SQL code.  
----

'''

OWNER_TAG = "owner"
GLUE_STUDIO_NOTEBOOK_IDENTIFIER = "GlueStudioNotebook"
SAGEMAKER_STUDIO_NOTEBOOK_IDENTIFIERS = ['SageMakerStudioPySparkNotebook', 'SageMakerStudioSparkNotebook']

# GlueStudio Env Variables
REQUEST_ORIGIN = "request_origin"
REGION = "region"
SESSION_ID = "session_id"
GLUE_ROLE_ARN = "glue_role_arn"

# Configs. These are the variable names which are used in the config file.
GLUE_VERSION = "glue_version"
GLUE_JOB_TYPE = "glue_job_type"
GLUE_TAGS = "aws_glue_tags"
USER_ID = "userId"
GLUE_PREFIX = "glue_"
ENDPOINT_URL = "endpoint_url"
PROFILE = "profile"
IAM_ROLE = "iam_role"
SESSION_ID_PREFIX = "session_id_prefix"
IDLE_TIMEOUT = "idle_timeout"
EXTRA_PY_FILES = "extra_py_files"
ENABLE_DATA_CATALOG = "enable_data_catalog"
EXTRA_JARS = "extra_jars"
ADDITIONAL_PYTHON_MODULES = "additional_python_modules"
CONNECTIONS = "connections"
NUMBER_OF_WORKERS = "number_of_workers"
MAX_CAPACITY = "max_capacity"
WORKER_TYPE = "worker_type"
SECURITY_CONFIG = "security_config"
SESSION_TYPE = "session_type"
SPARK_CONF = "spark_conf"
TAGS = "tags"
TEMP_CREDENTIALS = "temp_credentials"

REMAP_DEFAULT_ARGUMENTS = {
    "min_workers": "auto-scaling-ray-min-workers",
    "object_memory_head": "object_store_memory_head",
    "object_memory_worker": "object_store_memory_worker"
}

RAY_INVALID_REGION_ERROR = "An error occurred (InvalidInputException) when calling the CreateSession operation: Account not allowed to submit Glue Ray session"

# Retriever names
AWS_CONFIG_RETRIEVER = "AWSFileConfigRetriever"
DEFAULT_RETRIEVER = "DefaultRetriever"
OVERRIDES_RETRIEVER = "OverridenConfigRetriever"
SYSTEM_ENV_RETRIEVER = "SystemEnvConfigRetriever"


# Session Types
GLUE_ETL = "glueetl"
GLUE_RAY = "glueray"
GLUE_STREAMING = "gluestreaming"


from __future__ import print_function
import re
import ast
import boto3
import botocore
from botocore.exceptions import ClientError
from IPython.core.display import display_markdown
from IPython.core.magic import Magics, cell_magic, line_magic, magics_class

from .GlueSessionsConstants import *


@magics_class
class KernelMagics(Magics):

    def __init__(self, shell, data, kernel):
        super(KernelMagics, self).__init__(shell)
        self.data = data
        self.kernel = kernel

    @line_magic('iam_role')
    def set_glue_role_arn(self, glue_role_arn):
        self._validate_magic()
        glue_role_arn = self._strip_quotes(glue_role_arn)
        self.kernel._send_output(f'Current iam_role is {self.kernel.get_glue_role_arn()}')
        self.kernel.set_glue_role_arn(glue_role_arn)
        self.kernel._send_output(f'iam_role has been set to {glue_role_arn}.')

    @line_magic('idle_timeout')
    def set_idle_timeout(self, idle_timeout=2880):
        self._validate_magic()
        idle_timeout = self._strip_quotes(idle_timeout)
        self.kernel._send_output(f'Current idle_timeout is {self.kernel.get_idle_timeout()} minutes.')
        self.kernel.set_idle_timeout(int(idle_timeout))
        self.kernel._send_output(f'idle_timeout has been set to {self.kernel.get_idle_timeout()} minutes.')

    @line_magic('profile')
    def set_profile(self, profile):
        self._validate_magic()
        profile = self._strip_quotes(profile)
        self.kernel._send_output(f'Previous profile: {self.kernel.get_profile()}')
        self.kernel._send_output(f'Setting new profile to: {profile}')
        self.kernel.set_profile(profile)

    @line_magic('status')
    def get_status(self, line=None):
        if not self.kernel.get_session_id():
            self.kernel._send_output('There is no current session.')
            return
        status = self.kernel.get_current_session_status()
        role = self.kernel.get_current_session_role()
        session_id = self.kernel.get_current_session()['Id']
        created_on = self.kernel.get_current_session()['CreatedOn']
        glue_version = self.kernel.get_current_session()['GlueVersion']
        session_type = self.kernel.get_session_type()
        tags, exception_error = self.kernel.get_tags_from_resource()
        self.kernel._send_output(f'Session ID: {session_id}')
        self.kernel._send_output(f'Status: {status}')
        self.kernel._send_output(f'Role: {role}')
        self.kernel._send_output(f'CreatedOn: {created_on}')
        self.kernel._send_output(f'GlueVersion: {glue_version}')
        self.kernel._send_output(f'Session Type: {session_type}')
        if exception_error:
            self.kernel._send_output(exception_error)
        else:
            self.kernel._send_output(f'Tags: {tags}')
        # Print Max Capacity if it is set. Else print Number of Workers and Worker Type
        if self.kernel.get_max_capacity():
            self.kernel._send_output(f"Max Capacity: {self.kernel.get_max_capacity()}")
        else:
            self.kernel._send_output(f"Worker Type: {self.kernel.get_worker_type()}")
            self.kernel._send_output(f"Number of Workers: {self.kernel.get_number_of_workers()}")
        self.kernel._send_output(f'Region: {self.kernel.get_region()}')
        if self.kernel.get_connections():
            connections = self.kernel.get_connections()
            self.kernel._send_output(f"Connections: {connections}")
        default_arguments = self.kernel.get_default_arguments()
        args_list = [str(f"{k}: {v}") for k,v in default_arguments.items()]
        self.kernel._send_output(f"Arguments Passed: {args_list}")

    @line_magic('list_sessions')
    def list_sessions(self, line=None):
        sessions = self.kernel.get_sessions()
        if not sessions or not len(self.kernel.get_sessions().get('Ids')):
            self.kernel._send_output(f'There are no sessions to list.')
        else:
            ids = sessions.get('Ids')
            self.kernel._send_output(f'The first {len(ids)} sessions are:')
            session_table = SessionTable()

            for session in sessions.get('Sessions'):
                session_id = session['Id']
                status = session['Status']
                created_on = session['CreatedOn']
                session_table.add_session_info(session_id, status, created_on)

            table_output = session_table.format_session_info()
            self.kernel._send_output(table_output)

    @line_magic('stop_session')
    def stop_session(self, line=None):
        if not self.kernel.get_session_id():
            self.kernel._send_output('There is no current session.')
            return
        self.kernel.stop_session()
        self.kernel._send_output(f'Stopped session.')

    @line_magic('session_id')
    def set_session_id(self, line=None):
        if not self.kernel.get_session_id():
            self.kernel._send_output('There is no current session.')
        else:
            self.kernel._send_output(f'Current active Session ID: {self.kernel.get_session_id()}')

    @line_magic('session_id_prefix')
    def set_session_id_prefix(self, line=None):
        formatted_line = self._strip_quotes(line)
        if not self._validate_prefix(formatted_line):
            self.kernel._send_output(f'Not a valid input, the prefix only accepts -._#/A-Za-z0-9')
        else:
            self.kernel.set_session_id_prefix(formatted_line)
            self.kernel._send_output(f'Setting session ID prefix to {self.kernel.get_session_id_prefix()}')

    @line_magic('extra_py_files')
    def set_extra_py_files(self, line=None):
        if self.kernel.get_session_type() != SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            line = self._validate_list(line)
            if not line:
                return
            self.kernel._send_output("Extra py files to be included:")
            for s3_path in line.split(','):
                self.kernel._send_output(s3_path)
            self.kernel.set_extra_py_files(line)
        else:
            self.kernel._send_error_output(
                f"%extra_py_files is not supported in Glue Ray session type")

    @line_magic('additional_python_modules')
    def set_additional_python_modules(self, line=None):
        self._validate_magic()
        line = self._strip_quotes(line)
        line = self._validate_list(line)
        if not line:
            return
        self.kernel._send_output("Additional python modules to be included:")
        for module in line.split(','):
            self.kernel._send_output(module)
        self.kernel.set_additional_python_modules(line)

    @line_magic('extra_jars')
    def set_extra_jars(self, line=None):
        if self.kernel.get_session_type() != SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            line = self._validate_list(line)
            if not line:
                return
            self.kernel._send_output("Extra jars to be included:")
            for s3_path in line.split(','):
                self.kernel._send_output(s3_path)
            print(line)
            self.kernel.set_extra_jars(line)
        else:
            self.kernel._send_error_output(
                f"%extra_jars is not supported in Glue Ray session type")

    @line_magic('connections')
    def set_connections(self, line=None):
        if self.kernel.get_session_type() != SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            line = self._validate_list(line)
            if not line:
                return
            self.kernel._send_output("Connections to be included:")
            for connection in line.split(','):
                self.kernel._send_output(connection)
            self.kernel.set_connections(line)
        else:
            self.kernel._send_error_output(
                f"%connections is not supported in Glue Ray session type")

    @line_magic('glue_version')
    def set_glue_version(self,line=None):
        self._validate_magic()
        line = self._strip_quotes(line)
        self.kernel._send_output(f"Setting Glue version to: {line}")
        self.kernel.set_glue_version(line)

    @line_magic('endpoint')
    def set_endpoint(self, line=None):
        self._validate_magic()
        line = self._strip_quotes(line)
        previous_endpoint = self.kernel.get_endpoint_url()
        self.kernel._send_output(f'Previous endpoint: {previous_endpoint}')
        self.kernel._send_output(f'Setting new endpoint to: {line}')
        self.kernel.set_endpoint_url(line)
        self.kernel._send_output(f'Endpoint is set to: {line}')

    @line_magic('region')
    def set_region(self, line=None):
        self._validate_magic()
        line = self._strip_quotes(line)
        line = line.lower()
        previous_region = self.kernel.get_region()
        self.kernel._send_output(f'Previous region: {previous_region}')
        self.kernel._send_output(f'Setting new region to: {line}')
        self.kernel.set_region(line)
        self.kernel._send_output(f'Region is set to: {line}')

    @line_magic('number_of_workers')
    def set_number_of_workers(self, line=None):
        self._validate_magic()
        line = self._strip_quotes(line)
        self.kernel._send_output(f'Previous number of workers: {self.kernel.get_number_of_workers()}')
        self.kernel._send_output(f'Setting new number of workers to: {int(line)}')
        self.kernel.set_number_of_workers(line)

    @line_magic('worker_type')
    def set_worker_type(self, line=None):
        self._validate_magic()
        line = self._strip_quotes(line)
        self.kernel._send_output(f'Previous worker type: {self.kernel.get_worker_type()}')
        self.kernel._send_output(f'Setting new worker type to: {line}')
        self.kernel.set_worker_type(line)

    @line_magic('streaming')
    def set_streaming(self, line=None):
        self._validate_magic()
        self.kernel._send_output(f'Previous session type: {self.kernel.get_session_type()}')
        self.kernel._send_output('Setting new session type to Streaming')
        self.kernel.set_session_type(SessionType.streaming.pretty_name())

    @line_magic('etl')
    def set_etl(self, line=None):
        self._validate_magic()
        self.kernel._send_output(f'Previous session type: {self.kernel.get_session_type()}')
        self.kernel._send_output('Setting new session type to ETL')
        self.kernel.set_session_type(SessionType.etl.pretty_name())

    @line_magic('glue_ray')
    def set_glue_ray(self, line=None):
        self._validate_magic()
        self.kernel._send_output(f'Previous session type: {self.kernel.get_session_type()}')
        self.kernel._send_output(f'Setting new session type to {GLUE_RAY}')
        self.kernel.set_session_type(SessionType.glue_ray.pretty_name())

    @line_magic('security_config')
    def set_security_config(self, line=None):
        if self.kernel.get_session_type() != SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            self.kernel._send_output(f'Previous security_config: {self.kernel.get_security_config()}')
            self.kernel._send_output(f'Setting new security_config to: {line}')
            self.kernel.set_security_config(line)
        else:
            self.kernel._send_error_output(f"%security_config is not supported in Glue Ray session type")

    @line_magic('object_memory_head')
    def set_object_memory_head(self,line=None):
        if self.kernel.get_session_type() == SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            self.kernel._send_output(f'Setting new object_memory_head to: {int(line)} %')
            self.kernel.set_object_memory_head(line)
        else:
            self.kernel._send_error_output(f"%object_memory_head is only supported in Glue Ray session type")

    @line_magic('object_memory_worker')
    def set_object_memory_worker(self, line=None):
        if self.kernel.get_session_type() == SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            self.kernel._send_output(f'Setting new object_memory_worker to: {int(line)} %')
            self.kernel.set_object_memory_worker(line)
        else:
            self.kernel._send_error_output(f"%object_memory_worker is only supported in Glue Ray session type")

    @line_magic('min_workers')
    def set_min_workers(self, line=None):
        if self.kernel.get_session_type() == SessionType.glue_ray.pretty_name():
            self._validate_magic()
            line = self._strip_quotes(line)
            self.kernel._send_output(f'Setting new minimum workers: {int(line)}')
            self.kernel.set_min_workers(line)
        else:
            self.kernel._send_error_output(f"%min_workers is only supported in Glue Ray session type")

    @line_magic('session_type')
    def set_session_type(self, line=None):
        line = self._strip_quotes(line)
        self.kernel.set_session_type(line)

    @cell_magic('sql')
    def run_sql(self, line=None, cell=None):
        # No functionality here. SQL code formatted and passed by _handle_sql_code()
        # This function exists to declare the existence of the %%sql cell magic
        return

    @cell_magic('configure')
    def configure(self, line=None, cell=None):
        self._validate_magic()
        self.kernel.configure(cell)

    @line_magic('spark_conf')
    def spark_conf(self, conf=None):
        current_conf = self.kernel.get_spark_conf()
        if not conf:
            if not current_conf:
                self.kernel._send_output("No Spark configuration set")
            else:
                self.kernel._send_output(current_conf)
        else:
            self.kernel.set_spark_conf(conf)
            self.kernel._send_output(f'Previous Spark configuration: {current_conf}')
            self.kernel._send_output(f'Setting new Spark configuration to: {conf}')

    @line_magic('help')
    def help(self, line=None):
        self.kernel._send_markdown_output(HELP_TEXT)

    @cell_magic('tags')
    def add_tags_to_session(self, line=None, cell=None):
        session_status = self.kernel.get_session_status()
        if cell:
            try:
                cell = ast.literal_eval(cell)
                valid_tags = self._validate_dict(cell)
                if session_status == READY_SESSION_STATUS:
                    exception_error = self.kernel.add_tags_to_resource_ready_state(valid_tags)
                    if exception_error:
                        self.kernel._send_output(exception_error)
                        return
                    self.kernel.add_tag_to_resource(valid_tags)
                self.kernel._send_output(f'Tag {valid_tags} added')
            except Exception:
                tags = {"tag_key1": "tag_value1", 'tag_key2': 'tag_value2'}
                self.kernel._send_error_output(f"Please provide magics like {json.dumps(tags)}")
        return

    @cell_magic('assume_role')
    def assume_role_for_cross_account(self, line=None, cell=None):
        self._validate_magic()
        if cell:
            cell = ast.literal_eval(cell)
            try:
                if type(cell) is str:
                    response = boto3.client('sts').assume_role(RoleArn=cell, RoleSessionName=self.kernel.get_session_name())
                    self.kernel.set_temporary_credentials(response['Credentials']['AccessKeyId'], response['Credentials']['SecretAccessKey'], response['Credentials']['SessionToken'])
                elif type(cell) is dict:
                    valid_dict = self._validate_dict(cell)
                    self.kernel.set_temporary_credentials(valid_dict['aws_access_key_id'], valid_dict['aws_secret_access_key'], valid_dict['aws_session_token'])
                self.kernel._send_output('Role assumed successfully')
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                error_message = e.response['Error']['Message']
                if error_code in ('MalformedPolicyDocumentException', 'PackedPolicyTooLargeException', 'RegionDisabledException', 'ExpiredTokenException'):
                    self.kernel._send_error_output(f'Following exception encountered while assuming role for cross account user: {e} \n')
                    self.kernel._send_error_output(f'Error message: {error_message} \n')
                    self.kernel._print_traceback(e)
                else:
                    self.kernel._send_error_output(f'Error message: {error_message} \n')
                    self.kernel._print_traceback(e)
            except Exception as e:
                self.kernel._send_error_output(f"Not a valid input type {e}")
                self.kernel._print_traceback(e)

    def _strip_quotes(self, line):
        if not line:
            return None
        # Remove quotes
        line = line.strip('"')
        line = line.strip("'")
        return line

    def _validate_dict(self, cell):
        if not cell:
            return None
        else:
            return dict((key.strip(), value.strip()) for key, value in cell.items())

    def _validate_list(self, line):
        if not line:
            return None
        try:
            line = line.strip("[]")
            value_list = line.split(",")
            # create new list and strip leading and trailing spaces
            values = []
            for val in list(value_list):
                next_val = str(val).strip()
                if not next_val:
                    self.kernel._send_output("Empty values are not allowed to be passed.")
                    return None
                elif " " in next_val:
                    if "--" in next_val:
                        pass
                    else:
                        self.kernel._send_output(f'Invalid value. There is at least one blank space present in this value: {next_val}')
                        return None
                values.append(next_val)
        except Exception as e:
            self.kernel._send_output(f'Invalid list of inputs provided: {line}')
            return None
        return ",".join(values)

    def _validate_magic(self):
        session_id = self.kernel.get_session_id()
        if session_id:
            session = self.kernel.get_current_session()
            session_type = session['Command']['Name'] if session and not session == NOT_FOUND_SESSION_STATUS else None
            if session_type:
                self.kernel._send_error_output(f'You are already connected to a {session_type} session {session_id}.\n\nNo change will be made to the current session that is set as {session_type}. The session configuration change will apply to newly created sessions.')
            else:
                self.kernel._send_error_output(f'You are already connected to a session {session_id}.\n\nNo change will be made to the current session. The session configuration change will apply to newly created sessions.\n')

    def _validate_prefix(self, line):
        pattern = re.compile("[-/._#A-Za-z0-9]+")
        if not pattern.fullmatch(line):
            return False
        return True



from collections import defaultdict
from aws_glue_interactive_sessions_kernel.glue_kernel_utils.GlueSessionsConstants import SessionType


class Config:
    def __init__(self):
        self.profile = None
        self.endpoint_url = None
        self.region = None
        self.iam_role = None
        self.session_id_prefix = None
        self.max_capacity = None
        self.number_of_workers = None
        self.worker_type = None
        self.connections = None
        self.glue_version = None
        self.security_config = None
        self.idle_timeout = None
        self.tags = defaultdict()
        self.additional_python_modules = None
        self.extra_py_files = None
        self.enable_data_catalog = None
        self.extra_jars = None
        self.spark_conf = None
        self.session_type = None
        self.temp_credentials = None


default_configs = Config()
default_configs.session_type = SessionType.etl.pretty_name()
default_configs.glue_version = "4.0"




def convert_list_to_string(array_list):
    if not array_list:
        return None
    try:
        line = array_list.strip("[]")
        value_list = line.split(",")
        # create new list and strip leading and trailing spaces
        values = []
        for val in list(value_list):
            next_val = str(val).strip()
            if not next_val:
                return None
            elif " " in next_val:
                if "--" in next_val:
                    pass
                else:
                    return None
            values.append(next_val)
    except Exception as e:
        return None
    return ",".join(values)



import os

from setuptools import find_packages, setup

# Declare your non-python data files:
# Files underneath configuration/ will be copied into the build preserving the
# subdirectory structure if they exist.
data_files = []
for root, dirs, files in os.walk("configuration"):
    data_files.append(
        (os.path.relpath(root, "configuration"), [os.path.join(root, f) for f in files])
    )

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="aws-glue-sessions",
    version="1.0.2",
    description="Glue Interactive Sessions Jupyter kernel that integrates almost anywhere Jupyter does including your favorite IDEs.",
    url="https://aws.amazon.com/glue/",
    author="Glue Development Team",
    author_email="glue-sessions-preview@amazon.com",
    license="Apache 2.0 License",
    # declare your packages
    packages=find_packages(where="src", exclude=("test",)),
    package_dir={"": "src"},
    package_data={"": ["**/*.json", "**/*.sh", "**/*.png", "**/**/*.json"]},
    include_package_data=True,
    # include data files
    data_files=data_files,
    entry_points="""
        [console_scripts]
        install-glue-kernels = aws_glue_interactive_sessions_kernel.glue_kernel:install
    """,
    # Enable build-time format checking
    check_format=False,
    # Enable type checking
    test_mypy=False,
    # Enable linting at build time
    test_flake8=False,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "hdijupyterutils>=0.6",
        "autovizwidget>=0.6",
        "ipython>=4.0.2",
        "nose",
        "requests",
        "ipykernel>=6.12.1",
        "ipywidgets>5.0.0",
        "notebook>=4.2",
        "tornado>=4",
        "boto3>=1.21.31",
        "botocore>=1.24.24",
        "Click",
        "importlib_metadata>=4.11.3",
        "rich",
        "tabulate"
    ],
    license_files=['LICENSE']
)



from aws_glue_interactive_sessions_kernel.glue_kernel_configs.ConfigRetriever import ConfigRetriever


class DefaultConfigRetriever(ConfigRetriever):
    def __init__(self, default_configs):
        self.default_configs = default_configs

    def get_config_variable(self, config_name):
        return getattr(self.default_configs, config_name, None)



class GlueConfigManager:
    def __init__(self, config_retriever_ordered_dict):
        self.config_retriever_ordered_dict = config_retriever_ordered_dict

    def get_config_by_priority(self, config_name):
        for config_retriever in self.config_retriever_ordered_dict.values():
            value = config_retriever.get_config_variable(config_name)
            if value is not None:
                return value
        return None

    # This method works only for configs which support dict type as return type.
    # The values are iterated in reverse as the values with same keys can be replaced according to config priority.
    def get_config_from_all_retrievers(self, config_name):
        values = dict()
        for config_retriever in reversed(self.config_retriever_ordered_dict.values()):
            value = config_retriever.get_config_variable(config_name)
            if value is not None:
                values = {**values, **value}
        return values
